@@ -2,46 +2,59 @@ import math
 import builtins
 import re 
 
-from units_python.functions import fraction_decoder
-from units_python.constants import UNITS, SPECIAL_UNITS, PREFIXES
+## Todo:
+# Able to add units locally to a config file 
+# that will be loaded when units_python is imported
+
+from functions import fraction_decoder
+from constants import UNITS, SPECIAL_UNITS, TEN_EXPONENTS, SPECIAL_TEN_EXPONENTS
 
 ## Relevant constants ##
 pi = math.pi
 
 ## Relevant functions ##
-def sqrt(value, n=1): copy = value.copy(); copy.sqrt(n); return copy
-def nsqrt(value, n): copy = value.copy(); copy.sqrt(n); return copy
+def sqrt(value, n=2): return value.sqrt(n)
+def nsqrt(value, n): return value.sqrt(n)
 def round(value, digits): copy = value.copy(); copy.round(digits); return copy
 
 ## Value Class ##
 class v():
-    def __init__(self, value):
+    def __init__(self, value, ten_exponent = 0):
+        self.ten_exponent = ten_exponent
         value = value.replace("^", "**")
         self.nominators, self.denominators = fraction_decoder(value)
         self._get_value(value)
+        self._calibrate_ten_exponent()
 
+    def change_unit(self, str): self.unit = Unit(str)
     def copy(self): return v(self.__str__())
     def round(self, digits): self.value = builtins.round(self.value, digits)
-    def sqrt(self, n=1): return self.__pow__(1/n)
-    def __str__(self): return str(self.value) + " " + self.unit.get()
+    def sqrt(self, n=2): return self.__pow__(1/n)
+    def raw(self): return str(self.raw_value()) + " " +self.unit.get()
+    def raw_value(self): return self.value * 10 ** self.ten_exponent
+    def __str__(self): return str(self.value) + " " + self._get_ten_exponent() +self.unit.get()
     def __eq__(self, other): return str(self) == other
-    def __add__(self, other): return v(str(self.value + other.value) + " " + self.unit.get_add(other))
-    def __sub__(self, other): return v(str(self.value - other.value) + " " + self.unit.get_sub(other))
+    def __add__(self, other): return v(str(self.raw_value() + other.raw_value()) + " " + self.unit.get_add(other))
+    def __sub__(self, other): return v(str(self.raw_value() - other.raw_value()) + " " + self.unit.get_sub(other))
     def __rtruediv__(self, other): return self.__truediv__(other)
     def __rmul__(self, other): return self.__mul__(other)
-    def __pow__(self, other): return v(str(self.value ** other) + " " + self.unit ** other)
+    def __pow__(self, other): return v(str(self.raw_value() ** other) + " " + self.unit ** other)
 
     def __mul__(self, other): 
-        if isinstance(other, (int, float)): return v(str(self.value * other) + " " + self.unit.get())
+        if isinstance(other, (int, float)): return v(str(self.value * other) + " " + self.unit.get(), self.ten_exponent)
         value = self.value * other.value
         unit = self.unit * other.unit
-        return v(str(value) + " " + unit.get())
+        return v(str(value) + " " + unit.get(), ten_exponent = self.ten_exponent + other.ten_exponent)
 
     def __truediv__(self, other): 
-        if isinstance(other, (int, float)): return v(str(self.value / other) + " " + self.unit.get())
+        if isinstance(other, (int, float)): return v(str(self.value / other) + " " + self.unit.get(), self.ten_exponent)
         value = self.value / other.value
         unit = self.unit / other.unit
-        return v(str(value) + " " + unit.get())
+        return v(str(value) + " " + unit.get(), ten_exponent = self.ten_exponent - other.ten_exponent)
+
+    def _get_ten_exponent(self):
+        if self.ten_exponent != 0: return "* 10**" + str(self.ten_exponent) + " "
+        return " "
 
     def _get_value(self, value): 
         if len(self.nominators) and not self.denominators: self._get_single_value(value)
@@ -52,7 +65,7 @@ class v():
         nom_unit = Unit("")
         for nominator in self.nominators:
             if len(nominator.split()) == 1: 
-                nom_unit = self.handle_len_1_value(nominator, "mul", unit_obj = nom_unit)
+                nom_unit = self._handle_len_1_value(nominator, "mul", unit_obj = nom_unit)
             else: 
                 value_obj = v(nominator)
                 self.value *= value_obj.value
@@ -62,15 +75,15 @@ class v():
         if self.denominators:
             for denominator in self.denominators:
                 if len(denominator.split()) == 1: 
-                    denom_unit = self.handle_len_1_value(denominator, "div", unit_obj = denom_unit)
+                    denom_unit = self._handle_len_1_value(denominator, "div", unit_obj = denom_unit)
                 else: 
                     value_obj = v(denominator)
                     self.value /= value_obj.value
                     denom_unit *= value_obj.unit
         self.unit = nom_unit / denom_unit
 
-    def handle_len_1_value(self, value, method, unit_obj):
-        if self.is_hidden_float(value): 
+    def _handle_len_1_value(self, value, method, unit_obj):
+        if self._is_hidden_float(value): 
             if method == "mul": self.value *= value
             else: self.value /= value
         else: 
@@ -83,7 +96,7 @@ class v():
                 unit_obj *= Unit(value)
         return unit_obj
 
-    def is_hidden_float(self, value):
+    def _is_hidden_float(self, value):
         try: float(value); return True
         except: return False
 
@@ -93,6 +106,16 @@ class v():
         if len(split) == 1: self.value, self.unit = float(split[0]), Unit("")
         else: self.value, self.unit = float(split[0]), Unit(split[1])
         self.value *= self.unit.prefix
+        self.ten_exponent += self.unit.ten_exponent
+        self._calibrate_ten_exponent()
+
+    def _calibrate_ten_exponent(self):
+        while self.value > 10:
+            self.value /= 10
+            self.ten_exponent += 1
+        while self.value < 1:
+            self.value *= 10
+            self.ten_exponent -= 1
 
     def to(self, desired_unit):
         desired_unit_obj = v(f"1 {desired_unit}")
@@ -152,6 +175,7 @@ class Unit():
         self.set_unit(unit)
         self.set_powers()
         self.prefix = self.get_total_prefix()
+        self.ten_exponent = self.get_total_ten_exponent()
         self.unit = self.construct_unit()
 
     def set_unit(self, unit):
@@ -184,7 +208,7 @@ class Unit():
         if "**" in unit: unit, power = unit.split("**")
         else: power = 1
         return (unit, float(power))
-    
+
     def get_total_prefix(self):
         nominator_prefixes = [self.get_prefix(unit, power) for unit, power in zip(self.nominators, self.nominators_powers)]
         nominator_prefix = math.prod([prefix for (prefix, unit) in nominator_prefixes])
@@ -195,11 +219,26 @@ class Unit():
         return nominator_prefix / denominator_prefix
 
     def get_prefix(self, unit, power):
-        if unit[1:] in UNITS: prefix = PREFIXES[unit[0]]; unit = unit[1:]
-        elif unit in SPECIAL_UNITS: prefix, unit = SPECIAL_UNITS[unit]
+        if unit in SPECIAL_UNITS: prefix, unit = SPECIAL_UNITS[unit]
         else: prefix = 1
         prefix **= power
         return prefix, unit
+    
+    def get_total_ten_exponent(self):
+        nominator_prefixes = [self.get_ten_exponent(unit, power) for unit, power in zip(self.nominators, self.nominators_powers)]
+        nominator_prefix = sum([prefix for (prefix, unit) in nominator_prefixes])
+        self.nominators = [unit for (prefix, unit) in nominator_prefixes]
+        denominator_prefixes = [self.get_ten_exponent(unit, power) for unit, power in zip(self.denominators, self.denominators_powers)]
+        denominator_prefix = sum([prefix for (prefix, unit) in denominator_prefixes])
+        self.denominators = [unit for (prefix, unit) in denominator_prefixes]
+        return nominator_prefix - denominator_prefix
+
+    def get_ten_exponent(self, unit, power):
+        if unit[1:] in UNITS: ten_exponent = TEN_EXPONENTS[unit[0]]; unit = unit[1:]
+        elif unit in SPECIAL_TEN_EXPONENTS: ten_exponent, unit = SPECIAL_TEN_EXPONENTS[unit]
+        else: ten_exponent = 0
+        ten_exponent *= power
+        return ten_exponent, unit
 
     def simplify(self):
         nominator_dict, denominator_dict = self.get_simplify_dicts()
